# Don't Change

KILOMETERS_TO_MILES = 0.621371
DISTANCE_RADIUS = 6371
UNITS = ['mi', 'km']
OUTPUT = ['text', 'json']

# Constants defined by the World Geodetic System 1984 (WGS84)
A = 6378.137
B = 6356.7523142
ESQ = 6.69437999014 * 0.001

# Config

STORES_CSV = 'store-locations.csv'
DEFAULT_OUTPUT = 'text'
DEFAULT_UNITS = 'mi'
DEFAULT_ENCODING = 'utf-8-sig'
DEFAULT_DELIMITER = ','
DISTANCE_PRECISION = 2
INITIAL_RADIUS = 100
INC_RADIUS = 100

# CSV Field Names
STORE_FIELDS = {
    'NAME': 'Store Name',
    'LOCATION': 'Store Location',
    'COUNTY': 'County',
    'ADDRESS': 'Address',
    'CITY': 'City',
    'STATE': 'State',
    'ZIP_CODE': 'Zip Code',
    'LATITUDE': 'Latitude',
    'LONGITUDE': 'Longitude',
    'DISTANCE': 'Distance'
}



from storelocator.constants import (
    A,
    B,
    DEFAULT_UNITS,
    DISTANCE_PRECISION,
    DISTANCE_RADIUS,
    ESQ,
    KILOMETERS_TO_MILES,
    OUTPUT,
    STORE_FIELDS,
    UNITS
)
from decimal import (
    Decimal,
    ROUND_HALF_UP
)
import geocoder
import json
import math


def filter_stores(sp, lat_lng_ecef, initial_radius, inc_radius):
    """Queries StoresParser for stores near lat_lng_ecef within radius.

    Args:
        sp (obj): StoresParser instance.
        lat_lng_ecef (float): Converted ECEF lat/long.
        initial_radius (float): Initial search radius.
        inc_radius (float): Amount to increment search radius.
    Returns:
        List of nearby stores.

    """

    matches = []
    radius = initial_radius
    while len(matches) < 1:
        results = sp.query(lat_lng_ecef, radius)
        if results is not None:
            if len(results):
                matches.extend(results)
        radius += inc_radius
    return matches


def geodetic2ecef(lat, lon, alt=0):
    """Convert geodetic coordinates to ECEF.

    Args:
        lat (float): Latitude.
        lon (float): Longitude.
        alt (float, optional): Altitude.
    Returns:
        X, Y, Z ECEF coordinates.

    """

    lat, lon = math.radians(lat), math.radians(lon)
    xi = math.sqrt(1 - ESQ * math.sin(lat))
    x = (A / xi + alt) * math.cos(lat) * math.cos(lon)
    y = (A / xi + alt) * math.cos(lat) * math.sin(lon)
    z = (A / xi * (1 - ESQ) + alt) * math.sin(lat)
    return x, y, z


def euclidean_distance(distance):
    """Return the approximate Euclidean distance
    corresponding to the given great circle distance (in km).

    Args:
        distance (float): Distance (km).
    Returns:
        Approximate Euclidean distance.

    """

    return 2 * A * math.sin(distance / (2 * B))


def geocode(query):
    """Outputs latitude and longitude for a given address or zip code.

    Args:
        query (str, int): Address or zip code.
    Returns:
        Latitude and longitude (list of floats) or None.

    """

    # TODO(Austin) Add optional parameter for provider
    g = geocoder.google(query)
    return g.latlng


def validate_args(args):
    """Validates arguments and constructs query from them.

    Args:
        args (obj): Arguments object -> address, zip, units, output.
    Returns:
        {
            is_valid: (bool),
            query: (str/int or None)
        }

    """

    is_valid = True
    query = None
    if args.address is not None:
        if not isinstance(args.address, str):
            print('--address must be a string.')
            is_valid = False
        query = args.address
    elif args.zip is not None:
        if not (isinstance(args.zip, str) or isinstance(args.zip, int)):
            print('--zip must be a string or an integer.')
            is_valid = False
        query = args.zip
    else:
        print('--address or --zip must be specified.')
        is_valid = False
    if args.units is not None:
        if args.units not in UNITS:
            print('--units must be one of the following: {}'.format(UNITS))
            is_valid = False
    if args.output is not None:
        if args.output not in OUTPUT:
            print ('--output must be one of the following: {}'.format(OUTPUT))
            is_valid = False
    return {'is_valid': is_valid, 'query': query}


def format_distance(distance, precision=DISTANCE_PRECISION):
    """Formats float into rounded Decimal.

    Args:
        distance (float): Distance value to be formatted.
        precision (int, optional): Number of decimal places.
    Returns:
        Decimal representation of float rounded at given precision.

    """

    d = Decimal(distance)
    prec = '0.' + ''.join(
        ['0' if x < (precision - 1) else '1' for x in range(precision)]
    )
    return Decimal(d.quantize(Decimal(prec), rounding=ROUND_HALF_UP))


def format_result(result, distance, units, output):
    """Formats nearest store and corresponding distance into specified output.

    Args:
        result (dict or None): Result of nearest store.
        distance (float or None): Distance to nearest store.
        units (str): Distance metric (mi or km).
        output (str): Result format (text or json).
    Returns:
        Text or json representation of result and corresponding distance.

    """

    if output == 'text':
        if result is None or distance is None:
            return 'Unable to locate closest store.'
        formatted_result = (
            "Closest store is {} - {}, located in {} "
            "at {}, {}, {} {}. ({} {})"
            ).format(
            result[STORE_FIELDS['NAME']],
            result[STORE_FIELDS['LOCATION']],
            result[STORE_FIELDS['COUNTY']],
            result[STORE_FIELDS['ADDRESS']],
            result[STORE_FIELDS['CITY']],
            result[STORE_FIELDS['STATE']],
            result[STORE_FIELDS['ZIP_CODE']],
            format_distance(distance),
            units
        )
    else:
        if result is None or distance is None:
            print('Unable to locate closest store.')
            return {}
        result[STORE_FIELDS['DISTANCE']] = '{} {}'.format(
            format_distance(distance), units
        )
        formatted_result = json.dumps(result)
    return formatted_result


def calculate_distance(lat_lng_a, lat_lng_b, units=DEFAULT_UNITS):
    """Calculates distance in (mi or km) between two coords.

    Based on https://gist.github.com/rochacbruno/2883505

    Args:
        lat_lng_a (list(float)): Latitude and longitude of point A.
        lat_lng_b (list(float)): Latitude and longitude of point B.
        units (str, optional): Distance metric used for calculation (mi or km).
    Returns:
        Distance (float) in provided units (mi or km).

    """

    lat_a, lng_a = lat_lng_a
    lat_b, lng_b = lat_lng_b
    lat_diff = math.radians(lat_b - lat_a)
    lng_diff = math.radians(lng_b - lng_a)
    a = (
        math.sin(lat_diff / 2) * math.sin(lat_diff / 2) +
        math.cos(math.radians(lat_a)) * math.cos(math.radians(lat_b)) *
        math.sin(lng_diff / 2) * math.sin(lng_diff / 2)
        )
    c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
    distance = DISTANCE_RADIUS * c
    if units == 'mi':
        return distance * KILOMETERS_TO_MILES
    return distance


def find_nearest_store(lat_lng, stores, units):
    """Finds store from list of stores that is closest to a given set of coords.

    Args:
        lat_lng (list(float)): Latitude and longitude being compared to.
        stores (list(dict)): List of stores to search against.
        units (str): Distance metric used for comparison.
    Returns:
        Nearest store (dict or None) and corresponding distance (float or None).

    """
    result = None
    min_distance = None
    if lat_lng is not None and stores is not None:
        for store in stores:
            store_lat_lng = [
                float(
                    store[STORE_FIELDS['LATITUDE']]
                ),
                float(store[STORE_FIELDS['LONGITUDE']])
            ]
            store_distance = calculate_distance(lat_lng, store_lat_lng, units)
            if min_distance is None:
                min_distance = store_distance
                result = store
            else:
                if store_distance < min_distance:
                    min_distance = store_distance
                    result = store
    return result, min_distance



import codecs
from storelocator.constants import (
    DEFAULT_DELIMITER,
    DEFAULT_ENCODING,
    STORE_FIELDS
)
import csv
import numpy
import os
import pickle
from scipy.spatial import KDTree
from storelocator.util import (
    euclidean_distance,
    geodetic2ecef
)


class StoresParser(object):
    """StoresParser converts a CSV of stores data into a list of stores.

    """

    def __init__(
            self,
            file_path,
            encoding=DEFAULT_ENCODING,
            delimiter=DEFAULT_DELIMITER
            ):
        """Initialization creates a StoresParser instance to parse provided CSV.

        StoresParser is initialized with a relative file path to a CSV
        containing data about stores.  Optionally, the encoding of the CSV
        and the field delimiter may be passed to dictate how the CSV will
        be parsed.

        Args:
            file_path (str): Relative path to the CSV.
            encoding (str, optional): Encoding of the CSV.
            delimiter (str, optional): Field delimiter of the CSV.

        """
        self.file_path = file_path
        self.encoding = encoding
        self.delimiter = delimiter
        self.stores = None
        self.tree = None

    @staticmethod
    def get_StoresParser(stores_csv):
        if os.path.exists('{}.pkl'.format(stores_csv)):
            with open('{}.pkl'.format(stores_csv), 'rb') as input:
                sp = pickle.load(input)
        else:
            sp = StoresParser(stores_csv)
            sp.get_stores()
            sp.build_tree()
            sp.save()
        return sp

    @property
    def file_path(self):
        """Getter for file_path.

        """

        return self.__file_path

    @file_path.setter
    def file_path(self, fp):
        """Setter for file_path.

        """

        self.__file_path = fp

    @property
    def encoding(self):
        """Getter for encoding.

        """

        return self.__encoding

    @encoding.setter
    def encoding(self, enc):
        """Setter for encoding.

        """

        self.__encoding = enc

    @property
    def delimiter(self):
        """Getter for delimiter.

        """

        return self.__delimiter

    @delimiter.setter
    def delimiter(self, delim):
        """Setter for delimiter.

        """

        self.__delimiter = delim

    @property
    def stores(self):
        """Getter for stores.

        """

        return self.__stores

    @stores.setter
    def stores(self, stores):
        """Setter for stores.

        """

        self.__stores = stores

    @property
    def tree(self):
        """Getter for tree.

        """

        return self.__tree

    @tree.setter
    def tree(self, tree):
        """Setter for tree.

        """

        self.__tree = tree

    def save(self):
        """Saves StoresParser instance to .pkl file.

        """

        with open('{}.pkl'.format(self.file_path), 'wb') as output:
            pickle.dump(self, output, pickle.HIGHEST_PROTOCOL)

    def build_tree(self):
        """Creates KDTree representation of coordinate data.

        """

        if self.stores is not None:
            stores_ecef = []
            for store in self.stores:
                store_ecef = geodetic2ecef(
                    float(store[STORE_FIELDS['LATITUDE']]),
                    float(store[STORE_FIELDS['LONGITUDE']])
                )
                store['ecef'] = store_ecef
                stores_ecef.append(store_ecef)
            self.tree = KDTree(numpy.array(stores_ecef))

    def query(self, target_ecef, radius):
        """Searches for stores within a given radius of a location.

        Args:
            target_ecef (float): XYZ ECEF coords to search against.
            radius (float): Search radius.
        Returns:
            List of stores within a given radius of a location.

        """

        results = None
        if self.tree is not None:
            matches = self.tree.query_ball_point(
                [target_ecef],
                r=euclidean_distance(radius)
            )
            if len(matches):
                results = [self.stores[match] for match in matches[0]]
        return results

    def get_stores(self):
        """Parses stores data from CSV and returns a list of stores (dicts).

        The list of stores that is returned is sorted by latitude and then
        longitude.

        """

        with codecs.open(self.file_path, 'r', encoding=self.encoding) as f:
            reader = csv.DictReader(f, delimiter=self.delimiter)
            self.stores = list(reader)
        return self.stores



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="storelocator",
    version="1.0.1",
    author="Austin Brown",
    author_email="austinbrown34@gmail.com",
    description="Command-line tool to find nearest store.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/austinbrown34/StoreLocator",
    packages=setuptools.find_packages(),
    scripts=['scripts/find_store'],
    install_requires=[
        'geocoder==1.38.1',
        'nose2==0.7.4',
        'numpy==1.15.0',
        'scipy==1.1.0'
    ],
)
